#include "stdafx.h"
#include "general.h"
#include "materials.h"

#include "RaVi.h"

vector  RV_Target;
vector  RV_Up;

vector *RV_CamPtr;
vector *RV_RayPtr;
IntPar *RV_IntPtr;
vector *RV_OrgPtr;
vector *RV_DirPtr;
float  *RV_ShfPtr;
float  *RV_MDP;

#include "RaVi_Engine.h"

int       UseSSE=0;

BSPplane *RV_SceneTree;
Material  RV_Environment;
RV_PrP    RV_LtObj=NULL;

extern "C" {
  IntPar      Cmin;
  ushort      cframe;
  ushort      cray;
  RV_PrP      OrgOb;
  RV_PrP      CLtOb;
  RV_TriPara *bckdrp;
  LitPtr      RV_Lights=NULL;
  RGB_Color   Ambient = RGB_Color(0.4f, 0.4f, 0.4f);
}

float   ray_thresh;

void RV_Init()
{
  RV_CamPtr = (vector *)((((long)&RV_Data )&0xFFFFFFF0)     );
  RV_RayPtr = (vector *)((((long)&RV_Data )&0xFFFFFFF0)+0x10);
  RV_IntPtr = (IntPar *)((((long)&RV_Data )&0xFFFFFFF0)+0x20);
  RV_ShfPtr = (float  *)((((long)&RV_Data )&0xFFFFFFF0)+0x5C);
  RV_OrgPtr = (vector *)((((long)&RV_Data )&0xFFFFFFF0)+0x60);
  RV_DirPtr = (vector *)((((long)&RV_Data )&0xFFFFFFF0)+0x70);
    
  RV_MDP    = (float  *)((((long)&RV_MData)&0xFFFFFFF0)     );

  ((float *)(RV_CamPtr))[3]=1.0f;
  ((float *)(RV_RayPtr))[3]=1.0f;

  // check if Intel's streaming SIMD extensions are available
  UseSSE=CheckSSE();
  // if so, we are running on a Pentium III / Athlon XP or better.
  // That means we can use some handoptimized assembler routines
  // instead of their portable but slow C++ implementations.
  MatPtr mat=new Material; mat->next=NULL; mat->PBExp = 20.0f;
  mat->Reflection=RGB_Color(0.0f, 0.0f, 0.0f);
  mat->Specularity=0.0f; mat->type=0; mat->name = strdup("Backdrop");

  bckdrp = new RV_TriPara; bckdrp->type=1; bckdrp->flags=RV_CastNoShadow | RV_ReceiveNoShadow; bckdrp->Mat=mat;
  bckdrp->base=vector(0.0f, 0.0f, 0.0f);
  bckdrp->edg1=vector(1.0f, 0.0f, 0.0f);
  bckdrp->edg2=vector(0.0f, 1.0f, 0.0f);
  InitTP(bckdrp);
}

void RV_IntersectSimpleSphere0(RV_SimpleSphere *s)
{
  if(s->lastframe != cframe){
    s->pvec  = *RV_CamPtr - s->position;
    s->cterm = s->pvec.x*s->pvec.x + s->pvec.y*s->pvec.y + s->pvec.z*s->pvec.z - s->r2;
    s->lastframe = cframe;
  }
  float b = RV_RayPtr->x*s->pvec.x + RV_RayPtr->y*s->pvec.y + RV_RayPtr->z*s->pvec.z;
  float d = b*b-s->cterm;
  RV_IntPtr->t  = ((d>=0.0f)? -b - sqrtf(d) : -1.0f);
}

void RV_IntersectSimpleSphere1(RV_SimpleSphere *s, vector *Org, vector *Ray)
{
  vector pvc = *Org - s->position;
  float b = Ray->x*pvc.x + Ray->y*pvc.y + Ray->z*pvc.z;
  float c = pvc .x*pvc.x + pvc .y*pvc.y + pvc .z*pvc.z - s->r2;
  float d = b*b-c;
  if(d<=0.0f) { RV_IntPtr->t = -1.0f; return; }
  float t1 = -b - sqrtf(d);
  float t2 = -b + sqrtf(d);
  if(t1<t2){
    RV_IntPtr->t = ((t1>0.001f)? t1 : t2);
  } else {
    RV_IntPtr->t = ((t2>0.001f)? t2 : t1);
  }
}

void RV_IntersectComplexSphere1(RV_ComplexSphere *s, vector *Org, vector *Ray)
{
  vector pvc = *Org - s->position;
  float b = Ray->x*pvc.x + Ray->y*pvc.y + Ray->z*pvc.z;
  float c = pvc .x*pvc.x + pvc .y*pvc.y + pvc .z*pvc.z - s->r2;
  float d = b*b-c;
  if(d<=0.0f) { RV_IntPtr->t = -1.0f; return; }
  vector TRay=vector((*Ray).x*s->Tb.M11 + (*Ray).y*s->Tb.M21 + (*Ray).z*s->Tb.M31,
	                 (*Ray).x*s->Tb.M12 + (*Ray).y*s->Tb.M22 + (*Ray).z*s->Tb.M32,
	                 (*Ray).x*s->Tb.M13 + (*Ray).y*s->Tb.M23 + (*Ray).z*s->Tb.M33);
  vector TCam=(*Org) * s->Tb;
  float t1 = -b - sqrtf(d); vector IP1 = TCam+t1*TRay; float v1=ang(IP1.x, IP1.y); float u1=asinf(IP1.z);
  float t2 = -b + sqrtf(d); vector IP2 = TCam+t2*TRay; float v2=ang(IP2.x, IP2.y); float u2=asinf(IP2.z);
  if(s->vangle){
    if((u1 < s->ustart) || (u1 > s->ustop) || (v1 > s->vangle)){
      if((u2 < s->ustart) || (u2 > s->ustop) || (v2 > s->vangle)){
	    RV_IntPtr->t = -1.0f; return;
	  } else {
        RV_IntPtr->t=t2; RV_IntPtr->u=u2; RV_IntPtr->v=v2;
	  }
	} else {
	  if((u2 < s->ustart) || (u2 > s->ustop) || (v2 > s->vangle)){
        RV_IntPtr->t=t1; RV_IntPtr->u=u1; RV_IntPtr->v=v1;
	  } else {
	    if(t1<t2){
	      if(t1>0.001){
            RV_IntPtr->t=t1; RV_IntPtr->u=u1; RV_IntPtr->v=v1;
		  } else {
            RV_IntPtr->t=t2; RV_IntPtr->u=u2; RV_IntPtr->v=v2;
		  }
		} else {
	      if(t2>0.001){
            RV_IntPtr->t=t2; RV_IntPtr->u=u2; RV_IntPtr->v=v2;
		  } else {
            RV_IntPtr->t=t1; RV_IntPtr->u=u1; RV_IntPtr->v=v1;
		  }
		}
	  }
	}
  } else {
	if(t1<t2){
	  if(t1>0.001){
        RV_IntPtr->t=t1; RV_IntPtr->u=u1; RV_IntPtr->v=v1;
	  } else {
        RV_IntPtr->t=t2; RV_IntPtr->u=u2; RV_IntPtr->v=v2;
	  }
	} else {
	  if(t2>0.001){
        RV_IntPtr->t=t2; RV_IntPtr->u=u2; RV_IntPtr->v=v2;
	  } else {
        RV_IntPtr->t=t1; RV_IntPtr->u=u1; RV_IntPtr->v=v1;
	  }
	}
  }
}

void RV_IntersectComplexSphere0(RV_ComplexSphere *s) { RV_IntersectComplexSphere1(s, RV_CamPtr, RV_RayPtr); }

void RV_IntersectEllipsoid1(RV_Ellipsoid *e, vector *Org, vector *Ray)
{
  vector TRay=vector((*Ray).x*e->Tb.M11 + (*Ray).y*e->Tb.M21 + (*Ray).z*e->Tb.M31,
	                 (*Ray).x*e->Tb.M12 + (*Ray).y*e->Tb.M22 + (*Ray).z*e->Tb.M32,
	                 (*Ray).x*e->Tb.M13 + (*Ray).y*e->Tb.M23 + (*Ray).z*e->Tb.M33);
  vector TCam=(*Org) * e->Tb;
  float a = TRay.x*TRay.x + TRay.y*TRay.y + TRay.z*TRay.z;
  float b = TRay.x*TCam.x + TRay.y*TCam.y + TRay.z*TCam.z;
  float c = TCam.x*TCam.x + TCam.y*TCam.y + TCam.z*TCam.z-1.0f;
  float d = b*b-a*c;
  if(d<=0.0f){ RV_IntPtr->t=-1.0f; return; }
  float t1 = (-b - sqrtf(d))/a; vector IP1 = TCam + t1 * TRay; float v1=ang(IP1.x, IP1.y); float u1=asinf(IP1.z);
  float t2 = (-b + sqrtf(d))/a; vector IP2 = TCam + t2 * TRay; float v2=ang(IP2.x, IP2.y); float u2=asinf(IP2.z);
  if((u1 < e->ustart) || (u1 > e->ustop) || (e->vangle && (v1 > e->vangle))){
    if((u2 < e->ustart) || (u2 > e->ustop) || (e->vangle && (v2 > e->vangle))){
	  RV_IntPtr->t=-1.0f; return;
	} else {
      RV_IntPtr->t=t2; RV_IntPtr->IP=IP2; RV_IntPtr->u=u2; RV_IntPtr->v=v2;
	}
  } else {
    if((u2 < e->ustart) || (u2 > e->ustop) || (e->vangle && (v2 > e->vangle))){
      RV_IntPtr->t=t1; RV_IntPtr->IP=IP1; RV_IntPtr->u=u1; RV_IntPtr->v=v1;
	} else {
	  if(t1<t2){
	    if(t1>0.001f){
          RV_IntPtr->t=t1; RV_IntPtr->IP=IP1; RV_IntPtr->u=u1; RV_IntPtr->v=v1;
		} else {
          RV_IntPtr->t=t2; RV_IntPtr->IP=IP2; RV_IntPtr->u=u2; RV_IntPtr->v=v2;
		}
	  } else {
	    if(t2>0.001f){
          RV_IntPtr->t=t2; RV_IntPtr->IP=IP2; RV_IntPtr->u=u2; RV_IntPtr->v=v2;
		} else {
          RV_IntPtr->t=t1; RV_IntPtr->IP=IP1; RV_IntPtr->u=u1; RV_IntPtr->v=v1;
		}
	  }
	}
  }
}

void RV_IntersectEllipsoid0(RV_Ellipsoid *e) { RV_IntersectEllipsoid1(e, RV_CamPtr, RV_RayPtr); }

void RV_IntersectTriPara0(RV_TriPara *p)
{
  float RN=(*RV_RayPtr)%(p->N);
  if(RN && (!(p->flags & RV_SingleSided) || (RN<0.0f))){
    if(p->lastframe != cframe){
	   p->cpos       = (p->base - *RV_CamPtr)%(p->N);
       p->lastframe  = cframe;
	}
    RV_IntPtr->t = p->cpos/RN;
    if(RV_IntPtr->t>0.0f){
      RV_IntPtr->IP = *RV_CamPtr + RV_IntPtr->t * *RV_RayPtr;
      switch(p->algn){
	    case 0: RV_IntPtr->v = ((RV_IntPtr->IP.y - p->base.y)              *p->edg1.x - (RV_IntPtr->IP.x - p->base.x)*p->edg1.y) * p->deno; if(RV_IntPtr->v<0.0f){ RV_IntPtr->t=-1.0f; return; }
                RV_IntPtr->u =  (RV_IntPtr->IP.x - p->base.x - RV_IntPtr->v*p->edg2.x) / p->edg1.x;
                break;
        case 1: RV_IntPtr->v = ((RV_IntPtr->IP.z - p->base.z)              *p->edg1.y - (RV_IntPtr->IP.y - p->base.y)*p->edg1.z) * p->deno; if(RV_IntPtr->v<0.0f){ RV_IntPtr->t=-1.0f; return; }
                RV_IntPtr->u =  (RV_IntPtr->IP.y - p->base.y - RV_IntPtr->v*p->edg2.y) / p->edg1.y;
                break;
        case 2: RV_IntPtr->v = ((RV_IntPtr->IP.z - p->base.z)              *p->edg1.x - (RV_IntPtr->IP.x - p->base.x)*p->edg1.z) * p->deno; if(RV_IntPtr->v<0.0f){ RV_IntPtr->t=-1.0f; return; }
                RV_IntPtr->u =  (RV_IntPtr->IP.x - p->base.x - RV_IntPtr->v*p->edg2.x) / p->edg1.x;
                break;
        case 3: RV_IntPtr->v = ((RV_IntPtr->IP.x - p->base.x)              *p->edg1.y - (RV_IntPtr->IP.y - p->base.y)*p->edg1.x) * p->deno; if(RV_IntPtr->v<0.0f){ RV_IntPtr->t=-1.0f; return; }
                RV_IntPtr->u =  (RV_IntPtr->IP.y - p->base.y - RV_IntPtr->v*p->edg2.y) / p->edg1.y;
                break;
        case 4: RV_IntPtr->v = ((RV_IntPtr->IP.y - p->base.y)              *p->edg1.z - (RV_IntPtr->IP.z - p->base.z)*p->edg1.y) * p->deno; if(RV_IntPtr->v<0.0f){ RV_IntPtr->t=-1.0f; return; }
                RV_IntPtr->u =  (RV_IntPtr->IP.z - p->base.z - RV_IntPtr->v*p->edg2.z) / p->edg1.z;
                break;
        case 5: RV_IntPtr->v = ((RV_IntPtr->IP.x - p->base.x)              *p->edg1.z - (RV_IntPtr->IP.z - p->base.z)*p->edg1.x) * p->deno; if(RV_IntPtr->v<0.0f){ RV_IntPtr->t=-1.0f; return; }
                RV_IntPtr->u =  (RV_IntPtr->IP.z - p->base.z - RV_IntPtr->v*p->edg2.z) / p->edg1.z;
                break;
	  }
      if(RV_IntPtr->u>=0.0f){
        if(!(p->type)){
          if((RV_IntPtr->u       +   RV_IntPtr->v)>1.0f ) RV_IntPtr->t=-1.0f;
		} else {
          if((RV_IntPtr->u>1.0f) || (RV_IntPtr->v >1.0f)) RV_IntPtr->t=-1.0f;
		}
	  } else RV_IntPtr->t=-1.0f;
	}
  } else RV_IntPtr->t= -1.0f;
}

void RV_IntersectTriPara1(RV_TriPara *p, vector *Org, vector *Ray)
{
  float RN = *Ray % (p->N);
  if(RN && (!(p->flags & RV_SingleSided) || (RN<0.0f))){
    RV_IntPtr->t = ((p->base - *Org)%(p->N))/RN;
    if(RV_IntPtr->t>0.0f){
      RV_IntPtr->IP = *Org + RV_IntPtr->t * *Ray;
      switch(p->algn){
	    case 0: RV_IntPtr->v = ((RV_IntPtr->IP.y - p->base.y)              *p->edg1.x - (RV_IntPtr->IP.x - p->base.x)*p->edg1.y) * p->deno; if(RV_IntPtr->v<0.0f){ RV_IntPtr->t=-1.0f; return; }
                RV_IntPtr->u =  (RV_IntPtr->IP.x - p->base.x - RV_IntPtr->v*p->edg2.x) / p->edg1.x;
                break;
        case 1: RV_IntPtr->v = ((RV_IntPtr->IP.z - p->base.z)              *p->edg1.y - (RV_IntPtr->IP.y - p->base.y)*p->edg1.z) * p->deno; if(RV_IntPtr->v<0.0f){ RV_IntPtr->t=-1.0f; return; }
                RV_IntPtr->u =  (RV_IntPtr->IP.y - p->base.y - RV_IntPtr->v*p->edg2.y) / p->edg1.y;
                break;
        case 2: RV_IntPtr->v = ((RV_IntPtr->IP.z - p->base.z)              *p->edg1.x - (RV_IntPtr->IP.x - p->base.x)*p->edg1.z) * p->deno; if(RV_IntPtr->v<0.0f){ RV_IntPtr->t=-1.0f; return; }
                RV_IntPtr->u =  (RV_IntPtr->IP.x - p->base.x - RV_IntPtr->v*p->edg2.x) / p->edg1.x;
                break;
        case 3: RV_IntPtr->v = ((RV_IntPtr->IP.x - p->base.x)              *p->edg1.y - (RV_IntPtr->IP.y - p->base.y)*p->edg1.x) * p->deno; if(RV_IntPtr->v<0.0f){ RV_IntPtr->t=-1.0f; return; }
                RV_IntPtr->u =  (RV_IntPtr->IP.y - p->base.y - RV_IntPtr->v*p->edg2.y) / p->edg1.y;
                break;
        case 4: RV_IntPtr->v = ((RV_IntPtr->IP.y - p->base.y)              *p->edg1.z - (RV_IntPtr->IP.z - p->base.z)*p->edg1.y) * p->deno; if(RV_IntPtr->v<0.0f){ RV_IntPtr->t=-1.0f; return; }
                RV_IntPtr->u =  (RV_IntPtr->IP.z - p->base.z - RV_IntPtr->v*p->edg2.z) / p->edg1.z;
                break;
        case 5: RV_IntPtr->v = ((RV_IntPtr->IP.x - p->base.x)              *p->edg1.z - (RV_IntPtr->IP.z - p->base.z)*p->edg1.x) * p->deno; if(RV_IntPtr->v<0.0f){ RV_IntPtr->t=-1.0f; return; }
                RV_IntPtr->u =  (RV_IntPtr->IP.z - p->base.z - RV_IntPtr->v*p->edg2.z) / p->edg1.z;
                break;
	  }
      if(RV_IntPtr->u>=0.0f){
        if(!(p->type)){
          if((RV_IntPtr->u       +   RV_IntPtr->v)>1.0f ) RV_IntPtr->t=-1.0f;
		} else {
          if((RV_IntPtr->u>1.0f) || (RV_IntPtr->v >1.0f)) RV_IntPtr->t=-1.0f;
		}
	  } else RV_IntPtr->t=-1.0f;
	}
  } else RV_IntPtr->t= -1.0f;
}

void RV_IntersectCylinder0(RV_Cylinder *c)
{
  vector TRay=vector((*RV_RayPtr).x*c->Tb.M11 + (*RV_RayPtr).y*c->Tb.M21 + (*RV_RayPtr).z*c->Tb.M31,
                     (*RV_RayPtr).x*c->Tb.M12 + (*RV_RayPtr).y*c->Tb.M22 + (*RV_RayPtr).z*c->Tb.M32,
                     (*RV_RayPtr).x*c->Tb.M13 + (*RV_RayPtr).y*c->Tb.M23 + (*RV_RayPtr).z*c->Tb.M33);
  if(c->lastframe != cframe){
    c->TCam  = (*RV_CamPtr) * c->Tb;
    if(c->r_top){
      c->cterm = c->TCam.x*c->TCam.x + c->TCam.y*c->TCam.y - sqr(1.0f - c->r_top*c->TCam.z);
    } else {
      c->cterm = c->TCam.x*c->TCam.x + c->TCam.y*c->TCam.y -     1.0f;
    }
    c->lastframe = cframe;
  }
  float a,b,d;
  if(c->r_top){
    a = TRay.x*   TRay.x + TRay.y*   TRay.y - sqr             (       c->r_top*   TRay.z);
    b = TRay.x*c->TCam.x + TRay.y*c->TCam.y + c->r_top*TRay.z*(1.0f - c->r_top*c->TCam.z);
  } else {
    a = TRay.x*   TRay.x + TRay.y*   TRay.y;
    b = TRay.x*c->TCam.x + TRay.y*c->TCam.y;
  }
  d = b*b-a*c->cterm;
  if(d<=0.0f){ RV_IntPtr->t=-1.0f; return; }
  if(c->angle || (c->flags & RV_Inverse)){
    float t1 = (-b - sqrtf(d))/a; vector IP1 = c->TCam + t1 * TRay; float v1=ang(IP1.x, IP1.y);
    float t2 = (-b + sqrtf(d))/a; vector IP2 = c->TCam + t2 * TRay; float v2=ang(IP2.x, IP2.y);
    if((IP1.z < 0.0f) || (IP1.z > 1.0f) || (c->angle && (v1 > c->angle))){
      if((IP2.z < 0.0f) || (IP2.z > 1.0f) || (c->angle && (v2 > c->angle))){
        RV_IntPtr->t=-1.0f; return;
      } else {
        RV_IntPtr->t=t2; RV_IntPtr->IP=IP2; RV_IntPtr->u=v2; RV_IntPtr->v=IP2.z;
      }
    } else {
      if((IP2.z < 0.0f) || (IP2.z > 1.0f) || (c->angle && (v2 > c->angle))){
        RV_IntPtr->t=t1; RV_IntPtr->IP=IP1; RV_IntPtr->u=v1; RV_IntPtr->v=IP1.z;
      } else {
        if(t1<t2){
          if(t1>0.001f){
            RV_IntPtr->t=t1; RV_IntPtr->IP=IP1; RV_IntPtr->u=v1; RV_IntPtr->v=IP1.z;
          } else {
            RV_IntPtr->t=t2; RV_IntPtr->IP=IP2; RV_IntPtr->u=v2; RV_IntPtr->v=IP2.z;
          }
        } else {
          if(t2>0.001f){
            RV_IntPtr->t=t2; RV_IntPtr->IP=IP2; RV_IntPtr->u=v2; RV_IntPtr->v=IP2.z;
          } else {
            RV_IntPtr->t=t1; RV_IntPtr->IP=IP1; RV_IntPtr->u=v1; RV_IntPtr->v=IP1.z;
          }
        }
      }
    }
  } else {
    RV_IntPtr->t  = (-b - sqrtf(d))/a;
    RV_IntPtr->IP.z = c->TCam.z + RV_IntPtr->t * TRay.z;
    if((RV_IntPtr->IP.z<0.0f) || (RV_IntPtr->IP.z>1.0f)){ RV_IntPtr->t=-1.0f; return; }
    RV_IntPtr->IP.x = c->TCam.x + RV_IntPtr->t * TRay.x;
    RV_IntPtr->IP.y = c->TCam.y + RV_IntPtr->t * TRay.y;
  }
}

void RV_IntersectCylinder1(RV_Cylinder *c, vector *Org, vector *Ray)
{
  vector TRay=vector((*Ray).x*c->Tb.M11 + (*Ray).y*c->Tb.M21 + (*Ray).z*c->Tb.M31,
	                 (*Ray).x*c->Tb.M12 + (*Ray).y*c->Tb.M22 + (*Ray).z*c->Tb.M32,
	                 (*Ray).x*c->Tb.M13 + (*Ray).y*c->Tb.M23 + (*Ray).z*c->Tb.M33);
  vector TCam=(*Org) * c->Tb;
  float a,b,e,d;
  if(c->r_top){
    a = TRay.x*TRay.x + TRay.y*TRay.y - sqr             (       c->r_top*TRay.z);
    b = TRay.x*TCam.x + TRay.y*TCam.y + c->r_top*TRay.z*(1.0f - c->r_top*TCam.z);
    e = TCam.x*TCam.x + TCam.y*TCam.y - sqr             (1.0f - c->r_top*TCam.z);
  } else {
    a = TRay.x*TRay.x + TRay.y*TRay.y;
    b = TRay.x*TCam.x + TRay.y*TCam.y;
    e = TCam.x*TCam.x + TCam.y*TCam.y - 1.0f;
  }
  d = b*b-a*e;
  if(d<=0.0f){ RV_IntPtr->t=-1.0f; return; }
  if(c->angle || (c->flags & RV_Inverse)){
    float t1 = (-b - sqrtf(d))/a; vector IP1 = TCam + t1 * TRay; float v1=ang(IP1.x, IP1.y);
    float t2 = (-b + sqrtf(d))/a; vector IP2 = TCam + t2 * TRay; float v2=ang(IP2.x, IP2.y);
    if((IP1.z < 0.0f) || (IP1.z > 1.0f) || (c->angle && (v1 > c->angle))){
      if((IP2.z < 0.0f) || (IP2.z > 1.0f) || (c->angle && (v2 > c->angle))){
	    RV_IntPtr->t=-1.0f; return;
	  } else {
        RV_IntPtr->t=t2; RV_IntPtr->IP=IP2; RV_IntPtr->u=v2; RV_IntPtr->v=IP2.z;
	  }
	} else {
      if((IP2.z < 0.0f) || (IP2.z > 1.0f) || (c->angle && (v2 > c->angle))){
        RV_IntPtr->t=t1; RV_IntPtr->IP=IP1; RV_IntPtr->u=v1; RV_IntPtr->v=IP1.z;
	  } else {
	    if(t1<t2){
	      if(t1>0.001f){
            RV_IntPtr->t=t1; RV_IntPtr->IP=IP1; RV_IntPtr->u=v1; RV_IntPtr->v=IP1.z;
		  } else {
            RV_IntPtr->t=t2; RV_IntPtr->IP=IP2; RV_IntPtr->u=v2; RV_IntPtr->v=IP2.z;
		  }
		} else {
	      if(t2>0.001f){
            RV_IntPtr->t=t2; RV_IntPtr->IP=IP2; RV_IntPtr->u=v2; RV_IntPtr->v=IP2.z;
		  } else {
            RV_IntPtr->t=t1; RV_IntPtr->IP=IP1; RV_IntPtr->u=v1; RV_IntPtr->v=IP1.z;
		  }
		}
	  }
	}
  } else {
    RV_IntPtr->t  = (-b - sqrtf(d))/a;
    RV_IntPtr->IP.z = TCam.z + RV_IntPtr->t * TRay.z;
    if((RV_IntPtr->IP.z<0.0f) || (RV_IntPtr->IP.z>1.0f)){ RV_IntPtr->t=-1.0f; return; }
    RV_IntPtr->IP.x = TCam.x + RV_IntPtr->t * TRay.x;
    RV_IntPtr->IP.y = TCam.y + RV_IntPtr->t * TRay.y;
  }
}

// A coefficient smaller than SMALL_ENOUGH is considered to be zero (0.0).
#define SMALL_ENOUGH 1.0e-10

#define TWO_M_PI_3  2.0943951023931954923084
#define FOUR_M_PI_3 4.1887902047863909846168

static int solve_cubic(double *x, double *y)
{
  double Q, R, Q3, R2, sQ, d, an, theta;
  double A2, a0, a1, a2, a3;

  a0 = x[0];
  if(a0==0.0){
	return(0);
  } else {
    if(a0 != 1.0){
      a1 = x[1] / a0;
      a2 = x[2] / a0;
      a3 = x[3] / a0;
    } else {
      a1 = x[1];
      a2 = x[2];
      a3 = x[3];
    }
  }

  A2 =  a1 *  a1;   Q = (A2  -  3.0 * a2) /  9.0;
  R  = (a1 * (A2 - 4.5 * a2) + 13.5 * a3) / 27.0;
  Q3 = Q  * Q * Q;
  R2 = R  * R;
  d  = Q3 - R2;
  an = a1 / 3.0;

  if(d>=0.0){
    // Three real roots.
    d = R / sqrt(Q3); theta = acos(d) / 3.0; sQ = -2.0 * sqrt(Q);
    y[0] = sQ * cos(theta              ) - an;
    y[1] = sQ * cos(theta +  TWO_M_PI_3) - an;
    y[2] = sQ * cos(theta + FOUR_M_PI_3) - an;
    return(3);
  } else {
    sQ = pow(sqrt(R2 - Q3) + fabs(R), 1.0 / 3.0);
    if(R < 0){
      y[0] =  (sQ + Q / sQ) - an;
    } else {
      y[0] = -(sQ + Q / sQ) - an;
    }
    return(1);
  }
}

static int solve_quartic(double *x, double *results)
{
  double cubic[4], roots[3];
  double c12, z, p, q, q1, q2, r, d1, d2;
  double c0, c1, c2, c3, c4;

  // Make sure the quartic has a leading coefficient of 1.0
  c0 = x[0];
  if(c0 != 1.0){
    c1 = x[1] / c0;
    c2 = x[2] / c0;
    c3 = x[3] / c0;
    c4 = x[4] / c0;
  } else {
    c1 = x[1];
    c2 = x[2];
    c3 = x[3];
    c4 = x[4];
  }

  // Compute the cubic resolvant
  c12 =               c1  * c1;
  p   = -0.37500000 * c12 + c2;
  q   =  0.12500000 * c12 * c1  - 0.5000 * c1  * c2 + c3;
  r   = -0.01171875 * c12 * c12 + 0.0625 * c12 * c2 - 0.25 * c1 * c3 + c4;

  cubic[0] =  1.0;
  cubic[1] = -0.5 * p;
  cubic[2] = -r;
  cubic[3] =  0.5 * r * p - 0.125 * q * q;

  int i = solve_cubic(cubic, roots);

  if(i>0) z = roots[0]; else return 0;

  d1 = 2.0 * z - p; if(d1 < 0.0) if(d1 > -SMALL_ENOUGH) d1 = 0.0; else return 0;

  if(d1 < SMALL_ENOUGH){
    d2 = z * z - r;
    if(d2 < 0.0) return 0;
    d2 = sqrt(d2);
  } else {
    d1 = sqrt(d1);
    d2 = 0.5 * q / d1;
  }

  // Set up useful values for the quadratic factors
  q1 =    d1 * d1;
  q2 = -0.25 * c1; i = 0;

  // Solve the first quadratic
  p = q1 - 4.0 * (z - d2);
  if(p==0.0){
    results[i++] = -0.5 * d1 - q2;
  } else {
    if(p > 0){
      p = sqrt(p);
      results[i++] = -0.5 * (d1 + p) + q2;
      results[i++] = -0.5 * (d1 - p) + q2;
    }
  }

  // Solve the second quadratic
  p = q1 - 4.0 * (z + d2);
  if(p==0.0){
    results[i++] = 0.5 * d1 - q2;
  } else {
    if(p > 0){
      p = sqrt(p);
      results[i++] = 0.5 * (d1 + p) + q2;
      results[i++] = 0.5 * (d1 - p) + q2;
    }
  }

  return i;
}

int insector(float a, float start, float stop)
{
  if(start>=0.0f){
	return ((a>=start) && (a<=stop));
  } else {
	if(stop>=0.0f){
	  return (((a-6.2831853f)>=start) || ( a            <=stop));
	} else {
	  return (((a-6.2831853f)>=start) && ((a-6.2831853f)<=stop));
	}
  }
}

void RV_IntersectTorus1(RV_Torus *t, vector *Org, vector *Ray)
{
  vector TRay=vector((*Ray).x*t->Tb.M11 + (*Ray).y*t->Tb.M21 + (*Ray).z*t->Tb.M31,
	                 (*Ray).x*t->Tb.M12 + (*Ray).y*t->Tb.M22 + (*Ray).z*t->Tb.M32,
	                 (*Ray).x*t->Tb.M13 + (*Ray).y*t->Tb.M23 + (*Ray).z*t->Tb.M33);
  vector TCam=(*Org) * t->Tb;

  double c[5], r[4], Depth[4]; float v1,v2; vector IP,N,N1,N2;

  float len=Betrag(TRay); TRay=unit(TRay);
  
  double r12  = sqr(t->R1);
  double r22  = sqr(t->R2);

  double Py2  = TCam.z * TCam.z;
  double Dy2  = TRay.z * TRay.z;
  double PDy2 = TCam.z * TRay.z;

  double k1   = TCam.x * TCam.x + TCam.y * TCam.y + Py2 - r12 - r22;
  double k2   = TCam.x * TRay.x + TCam.y * TRay.y + PDy2;

  c[0] = 1.0;
  c[1] = 4.0 *  k2;
  c[2] = 2.0 * (k1 + 2.0 * (k2 * k2 + r12 * Dy2));
  c[3] = 4.0 * (k2 * k1  + 2.0 * r12 * PDy2);
  c[4] =  k1 *  k1 + 4.0 * r12 * (Py2 - r22);

  int n = solve_quartic(c, r); int i=0; while(n--) Depth[i++] = r[n] / len;
  
  if(i){
	doubsort(i,Depth);
    RV_IntPtr->t=(float)Depth[0]; IP = TCam + RV_IntPtr->t * TRay*len;
	v1 = ang(IP.x, IP.y); N1=IP-vector(t->R1*cosf(v1), t->R1*sinf(v1), 0.0f);
	v2 = v1+3.1415927f;   N2=IP-vector(t->R1*cosf(v2), t->R1*sinf(v2), 0.0f);
	if(fabs(Betrag(N1) - t->R2) < fabs(Betrag(N2) - t->R2)){ RV_IntPtr->v=v1; RV_IntPtr->N=unit(N1); } else { RV_IntPtr->v=v2; RV_IntPtr->N=unit(N2); }
	RV_IntPtr->u = IP.z<0.0f? 6.2831853f-(float)acs(RV_IntPtr->N % vector(cosf(RV_IntPtr->v), sinf(RV_IntPtr->v), 0.0f)) : (float)acs(RV_IntPtr->N % vector(cosf(RV_IntPtr->v), sinf(RV_IntPtr->v), 0.0f));
	if((RV_IntPtr->t < 0.001f) || (t->vangle && ((RV_IntPtr->v > t->vangle) || !insector(RV_IntPtr->u, t->ustart, t->ustop)))){
	  if(i<2) { RV_IntPtr->t = -1.0f; return; }
      RV_IntPtr->t=(float)Depth[1]; IP = TCam + RV_IntPtr->t * TRay*len;
	  v1 = ang(IP.x, IP.y); N1=IP-vector(t->R1*cosf(v1), t->R1*sinf(v1), 0.0f);
	  v2 = v1+3.1415927f;   N2=IP-vector(t->R1*cosf(v2), t->R1*sinf(v2), 0.0f);
	  if(fabs(Betrag(N1) - t->R2) < fabs(Betrag(N2) - t->R2)){ RV_IntPtr->v=v1; RV_IntPtr->N=unit(N1); } else { RV_IntPtr->v=v2; RV_IntPtr->N=unit(N2); }
	  RV_IntPtr->u = IP.z<0.0f? 6.2831853f-(float)acs(RV_IntPtr->N % vector(cosf(RV_IntPtr->v), sinf(RV_IntPtr->v), 0.0f)) : (float)acs(RV_IntPtr->N % vector(cosf(RV_IntPtr->v), sinf(RV_IntPtr->v), 0.0f));
	  if((RV_IntPtr->t < 0.001f) || (t->vangle && ((RV_IntPtr->v > t->vangle) || !insector(RV_IntPtr->u, t->ustart, t->ustop)))){
        if(i<3) { RV_IntPtr->t = -1.0f; return; }
        RV_IntPtr->t=(float)Depth[2]; IP = TCam + RV_IntPtr->t * TRay*len;
	    v1 = ang(IP.x, IP.y); N1=IP-vector(t->R1*cosf(v1), t->R1*sinf(v1), 0.0f);
	    v2 = v1+3.1415927f;   N2=IP-vector(t->R1*cosf(v2), t->R1*sinf(v2), 0.0f);
	    if(fabs(Betrag(N1) - t->R2) < fabs(Betrag(N2) - t->R2)){ RV_IntPtr->v=v1; RV_IntPtr->N=unit(N1); } else { RV_IntPtr->v=v2; RV_IntPtr->N=unit(N2); }
	    RV_IntPtr->u = IP.z<0.0f? 6.2831853f-(float)acs(RV_IntPtr->N % vector(cosf(RV_IntPtr->v), sinf(RV_IntPtr->v), 0.0f)) : (float)acs(RV_IntPtr->N % vector(cosf(RV_IntPtr->v), sinf(RV_IntPtr->v), 0.0f));
	    if((RV_IntPtr->t < 0.001f) || (t->vangle && ((RV_IntPtr->v > t->vangle) || !insector(RV_IntPtr->u, t->ustart, t->ustop)))){
          if(i<3) { RV_IntPtr->t = -1.0f; return; }
          RV_IntPtr->t=(float)Depth[3]; IP = TCam + RV_IntPtr->t * TRay*len;
	      v1 = ang(IP.x, IP.y); N1=IP-vector(t->R1*cosf(v1), t->R1*sinf(v1), 0.0f);
	      v2 = v1+3.1415927f;   N2=IP-vector(t->R1*cosf(v2), t->R1*sinf(v2), 0.0f);
	      if(fabs(Betrag(N1) - t->R2) < fabs(Betrag(N2) - t->R2)){ RV_IntPtr->v=v1; RV_IntPtr->N=unit(N1); } else { RV_IntPtr->v=v2; RV_IntPtr->N=unit(N2); }
	      RV_IntPtr->u = IP.z<0.0f? 6.2831853f-(float)acs(RV_IntPtr->N % vector(cosf(RV_IntPtr->v), sinf(RV_IntPtr->v), 0.0f)) : (float)acs(RV_IntPtr->N % vector(cosf(RV_IntPtr->v), sinf(RV_IntPtr->v), 0.0f));
	      if(t->vangle && ((RV_IntPtr->v > t->vangle) || !insector(RV_IntPtr->u, t->ustart, t->ustop))) { RV_IntPtr->t = -1.0f; return; }
	    }
	  }
	}
  } else RV_IntPtr->t = -1.0f;
}

void RV_IntersectTorus0(RV_Torus *c) { RV_IntersectTorus1(c, RV_CamPtr, RV_RayPtr); }

void RV_IntersectParaboloid1(RV_Paraboloid *p, vector *Org, vector *Ray)
{
  vector TRay=vector((*Ray).x*p->Tb.M11 + (*Ray).y*p->Tb.M21 + (*Ray).z*p->Tb.M31,
	                 (*Ray).x*p->Tb.M12 + (*Ray).y*p->Tb.M22 + (*Ray).z*p->Tb.M32,
	                 (*Ray).x*p->Tb.M13 + (*Ray).y*p->Tb.M23 + (*Ray).z*p->Tb.M33);
  vector TCam=(*Org) * p->Tb;
  float a = TRay.x*TRay.x + TRay.y*TRay.y;
  float b = TRay.x*TCam.x + TRay.y*TCam.y - 0.5f*TRay.z;
  float c = TCam.x*TCam.x + TCam.y*TCam.y -      TCam.z;
  float d = b*b-a*c;
  if(d<=0.0f){ RV_IntPtr->t=-1.0f; return; }
  float t1 = (-b - sqrtf(d))/a; vector IP1 = TCam + t1 * TRay; float v1=ang(IP1.x, IP1.y);
  float t2 = (-b + sqrtf(d))/a; vector IP2 = TCam + t2 * TRay; float v2=ang(IP2.x, IP2.y);
  if((IP1.z > p->height) || (p->angle && (v1 > p->angle))){
    if((IP2.z > p->height) || (p->angle && (v2 > p->angle))){
	  RV_IntPtr->t=-1.0f; return;
	} else {
      RV_IntPtr->t=t2; RV_IntPtr->IP=IP2; RV_IntPtr->v=v2;
	}
  } else {
    if((IP2.z > p->height) || (p->angle && (v2 > p->angle))){
      RV_IntPtr->t=t1; RV_IntPtr->IP=IP1; RV_IntPtr->v=v1;
	} else {
	  if(t1<t2){
	    if(t1>0.001f){
          RV_IntPtr->t=t1; RV_IntPtr->IP=IP1; RV_IntPtr->v=v1;
		} else {
          RV_IntPtr->t=t2; RV_IntPtr->IP=IP2; RV_IntPtr->v=v2;
		}
	  } else {
	    if(t2>0.001f){
          RV_IntPtr->t=t2; RV_IntPtr->IP=IP2; RV_IntPtr->v=v2;
		} else {
          RV_IntPtr->t=t1; RV_IntPtr->IP=IP1; RV_IntPtr->v=v1;
		}
	  }
	}
  }
  RV_IntPtr->u=RV_IntPtr->IP.z;
}

void RV_IntersectParaboloid0(RV_Paraboloid *p) { RV_IntersectParaboloid1(p, RV_CamPtr, RV_RayPtr); }

void RV_IntersectSimpleRing0(RV_SimpleRing *r)
{
  if(r->lastframe != cframe){
	 r->cpos       = (r->center-(*RV_CamPtr))%(r->N);
     r->lastframe  = cframe;
  }
  float RN=(*RV_RayPtr)%(r->N);
  if(RN){
    RV_IntPtr->t = r->cpos/RN;
    if(RV_IntPtr->t>0.0f){
      RV_IntPtr->IP = *RV_CamPtr + RV_IntPtr->t * *RV_RayPtr;
	  float r2=sqr(RV_IntPtr->IP.x-r->center.x)+sqr(RV_IntPtr->IP.y-r->center.y)+sqr(RV_IntPtr->IP.z-r->center.z);
	  if(r2>r->r_o) RV_IntPtr->t= -1.0f;
	} else RV_IntPtr->t= -1.0f;
  } else RV_IntPtr->t= -1.0f;
}

void RV_IntersectSimpleRing1(RV_SimpleRing *r, vector *Org, vector *Ray)
{
  float RN=(*Ray)%(r->N);
  if(RN){
    RV_IntPtr->t = ((r->center-(*Org))%(r->N))/RN;
    if(RV_IntPtr->t>0.0f){
      RV_IntPtr->IP = *Org + RV_IntPtr->t * *Ray;
	  float r2=sqr(RV_IntPtr->IP.x-r->center.x)+sqr(RV_IntPtr->IP.y-r->center.y)+sqr(RV_IntPtr->IP.z-r->center.z);
	  if((r2>r->r_o) || (r2<r->r_i)) RV_IntPtr->t= -1.0f;
	} else RV_IntPtr->t= -1.0f;
  } else RV_IntPtr->t= -1.0f;
}

void RV_IntersectComplexRing1(RV_ComplexRing *r, vector *Org, vector *Ray)
{
  float RN=(*Ray)%(r->N);
  if(RN){
    RV_IntPtr->t = ((r->center-(*Org))%(r->N))/RN;
    if(RV_IntPtr->t>0.0f){
      RV_IntPtr->IP = *Org + RV_IntPtr->t * *Ray;
	  RV_IntPtr->u  = sqr(RV_IntPtr->IP.x-r->center.x)+sqr(RV_IntPtr->IP.y-r->center.y)+sqr(RV_IntPtr->IP.z-r->center.z);
	  if((RV_IntPtr->u > r->r_o) || (RV_IntPtr->u < r->r_i)) { RV_IntPtr->t= -1.0f; return; }
	  vector OP = RV_IntPtr->IP * r->Tb; RV_IntPtr->v = ang(OP.x, OP.y);
	  if(r->angle && (RV_IntPtr->v > r->angle)) { RV_IntPtr->t= -1.0f; return; }
	} else RV_IntPtr->t= -1.0f;
  } else RV_IntPtr->t= -1.0f;
}

void RV_IntersectComplexRing0(RV_ComplexRing *r) { RV_IntersectComplexRing1(r, RV_CamPtr, RV_RayPtr); }

#define RIGHT	0
#define LEFT	1
#define MIDDLE	2

extern "C" {
  vector             BBMin,BBMax;
  vector             BBcandidatePlane;
  unsigned short int BBinside,BBquadrant[3];
  unsigned short int BBlastframe=42;
}

void RV_IntersectBoundingBox0c()
{
  int whichPlane;
  vector maxT;

  // Find candidate planes - only needed once per frame
  if(BBlastframe != cframe){
	BBlastframe  = cframe; BBinside=true;

    if(       RV_CamPtr->x < BBMin.x) {
      BBquadrant[0] = LEFT;  BBcandidatePlane.x = BBMin.x; BBinside=false;
	} else if(RV_CamPtr->x > BBMax.x) {
      BBquadrant[0] = RIGHT; BBcandidatePlane.x = BBMax.x; BBinside=false;
	} else {
      BBquadrant[0] = MIDDLE;
	}

    if(       RV_CamPtr->y < BBMin.y) {
      BBquadrant[1] = LEFT;  BBcandidatePlane.y = BBMin.y; BBinside=false;
	} else if(RV_CamPtr->y > BBMax.y) {
      BBquadrant[1] = RIGHT; BBcandidatePlane.y = BBMax.y; BBinside=false;
	} else {
      BBquadrant[1] = MIDDLE;
	}

    if(       RV_CamPtr->z < BBMin.z) {
      BBquadrant[2] = LEFT;  BBcandidatePlane.z = BBMin.z; BBinside=false;
	} else if(RV_CamPtr->z > BBMax.z) {
      BBquadrant[2] = RIGHT; BBcandidatePlane.z = BBMax.z; BBinside=false;
	} else {
      BBquadrant[2] = MIDDLE;
	}
  }

  if(BBinside){ RV_IntPtr->t = 1.0f; RV_IntPtr->IP = *RV_CamPtr; return; }

  // calculate distances to candidate planes
  if(BBquadrant[0] != MIDDLE && RV_RayPtr->x) maxT.x = (BBcandidatePlane.x - RV_CamPtr->x) / RV_RayPtr->x; else	maxT.x = -1.0f;
  if(BBquadrant[1] != MIDDLE && RV_RayPtr->y) maxT.y = (BBcandidatePlane.y - RV_CamPtr->y) / RV_RayPtr->y; else	maxT.y = -1.0f;
  if(BBquadrant[2] != MIDDLE && RV_RayPtr->z) maxT.z = (BBcandidatePlane.z - RV_CamPtr->z) / RV_RayPtr->z; else	maxT.z = -1.0f;

  // get largest of the maxT's for final choice of intersection
  whichPlane = 0; RV_IntPtr->t = maxT.x;
  if(RV_IntPtr->t < maxT.y){ whichPlane = 1; RV_IntPtr->t = maxT.y; }
  if(RV_IntPtr->t < maxT.z){ whichPlane = 2; RV_IntPtr->t = maxT.z; }

  // check final candidate actually inside box
  if(RV_IntPtr->t < 0.0f) return;

  if(whichPlane != 0) {
	RV_IntPtr->IP.x = RV_CamPtr->x + RV_IntPtr->t * RV_RayPtr->x;
    if(RV_IntPtr->IP.x < BBMin.x || RV_IntPtr->IP.x > BBMax.x){ RV_IntPtr->t = -1.0f; return; }
  } else {
    RV_IntPtr->IP.x = BBcandidatePlane.x;
  }
  if(whichPlane != 1) {
	RV_IntPtr->IP.y = RV_CamPtr->y + RV_IntPtr->t * RV_RayPtr->y;
    if(RV_IntPtr->IP.y < BBMin.y || RV_IntPtr->IP.y > BBMax.y){ RV_IntPtr->t = -1.0f; return; }
  } else {
    RV_IntPtr->IP.y = BBcandidatePlane.y;
  }
  if(whichPlane != 2) {
	RV_IntPtr->IP.z = RV_CamPtr->z + RV_IntPtr->t * RV_RayPtr->z;
    if(RV_IntPtr->IP.z < BBMin.z || RV_IntPtr->IP.z > BBMax.z){ RV_IntPtr->t = -1.0f; return; }
  } else {
    RV_IntPtr->IP.z = BBcandidatePlane.z;
  }
}

void RV_IntersectBoundingBox1()
{
  vector  maxT;
  vector  candidatePlane;
  int     whichPlane;
  char    quadrant[3];
  bool    inside=true;

  // Find candidate planes
  if(       RV_OrgPtr->x < BBMin.x) {
    quadrant[0] = LEFT;  candidatePlane.x = BBMin.x; inside=false;
  } else if(RV_OrgPtr->x > BBMax.x) {
    quadrant[0] = RIGHT; candidatePlane.x = BBMax.x; inside=false;
  } else {
    quadrant[0] = MIDDLE;
  }

  if(       RV_OrgPtr->y < BBMin.y) {
    quadrant[1] = LEFT;  candidatePlane.y = BBMin.y; inside=false;
  } else if(RV_OrgPtr->y > BBMax.y) {
    quadrant[1] = RIGHT; candidatePlane.y = BBMax.y; inside=false;
  } else {
    quadrant[1] = MIDDLE;
  }

  if(       RV_OrgPtr->z < BBMin.z) {
    quadrant[2] = LEFT;  candidatePlane.z = BBMin.z; inside=false;
  } else if(RV_OrgPtr->z > BBMax.z) {
    quadrant[2] = RIGHT; candidatePlane.z = BBMax.z; inside=false;
  } else {
    quadrant[2] = MIDDLE;
  }

  if(inside){ RV_IntPtr->t = 1.0f; RV_IntPtr->IP = *RV_OrgPtr; return; }

  // calculate distances to candidate planes
  if(quadrant[0] != MIDDLE && RV_DirPtr->x) maxT.x = (candidatePlane.x - RV_OrgPtr->x) / RV_DirPtr->x; else	maxT.x = -1.0f;
  if(quadrant[1] != MIDDLE && RV_DirPtr->y) maxT.y = (candidatePlane.y - RV_OrgPtr->y) / RV_DirPtr->y; else	maxT.y = -1.0f;
  if(quadrant[2] != MIDDLE && RV_DirPtr->z) maxT.z = (candidatePlane.z - RV_OrgPtr->z) / RV_DirPtr->z; else	maxT.z = -1.0f;

  // get largest of the maxT's for final choice of intersection
  whichPlane = 0; RV_IntPtr->t = maxT.x;
  if(RV_IntPtr->t < maxT.y){ whichPlane = 1; RV_IntPtr->t = maxT.y; }
  if(RV_IntPtr->t < maxT.z){ whichPlane = 2; RV_IntPtr->t = maxT.z; }

  // check final candidate actually inside box
  if(RV_IntPtr->t < 0.0f) return;

  if(whichPlane != 0) {
	RV_IntPtr->IP.x = RV_OrgPtr->x + RV_IntPtr->t * RV_DirPtr->x;
    if(RV_IntPtr->IP.x < BBMin.x || RV_IntPtr->IP.x > BBMax.x){ RV_IntPtr->t = -1.0f; return; }
  } else {
    RV_IntPtr->IP.x = candidatePlane.x;
  }
  if(whichPlane != 1) {
	RV_IntPtr->IP.y = RV_OrgPtr->y + RV_IntPtr->t * RV_DirPtr->y;
    if(RV_IntPtr->IP.y < BBMin.y || RV_IntPtr->IP.y > BBMax.y){ RV_IntPtr->t = -1.0f; return; }
  } else {
    RV_IntPtr->IP.y = candidatePlane.y;
  }
  if(whichPlane != 2) {
	RV_IntPtr->IP.z = RV_OrgPtr->z + RV_IntPtr->t * RV_DirPtr->z;
    if(RV_IntPtr->IP.z < BBMin.z || RV_IntPtr->IP.z > BBMax.z){ RV_IntPtr->t = -1.0f; return; }
  } else {
    RV_IntPtr->IP.z = candidatePlane.z;
  }
}

void RV_IntersectCube0(RV_Cube *c)
{
  if(c->lastframe != cframe){
	c->or[0]=((c->top  - *RV_CamPtr) %  c->axis1);
	c->or[1]=((c->base - *RV_CamPtr) % -c->axis1);
	c->or[2]=((c->top  - *RV_CamPtr) %  c->axis2);
	c->or[3]=((c->base - *RV_CamPtr) % -c->axis2);
	c->or[4]=((c->top  - *RV_CamPtr) %  c->axis3);
	c->or[5]=((c->base - *RV_CamPtr) % -c->axis3);
    c->lastframe = cframe;
  }
  float t[6];
  t[0] = c->or[0]/(*RV_RayPtr %  c->axis1);
  t[1] = c->or[1]/(*RV_RayPtr % -c->axis1);
  t[2] = c->or[2]/(*RV_RayPtr %  c->axis2);
  t[3] = c->or[3]/(*RV_RayPtr % -c->axis2);
  t[4] = c->or[4]/(*RV_RayPtr %  c->axis3);
  t[5] = c->or[5]/(*RV_RayPtr % -c->axis3);

  char x_near = ((t[0]<t[1])? 0 : 1);
  char y_near = ((t[2]<t[3])? 2 : 3);
  char z_near = ((t[4]<t[5])? 4 : 5);

  RV_IntPtr->i = ((t[x_near] > t[y_near])? x_near : y_near);
  if(t[z_near] > t[RV_IntPtr->i]) RV_IntPtr->i=z_near;

  if( t[RV_IntPtr->i] < min(min(t[1-x_near],t[5-y_near]),t[9-z_near]) ){
	RV_IntPtr->t = t[RV_IntPtr->i];
  } else {
	RV_IntPtr->t = -1.0f;
  }
}

void RV_IntersectCube1(RV_Cube *c, vector *Org, vector *Ray)
{
  float t[6];
  t[0] = ((c->top  - *Org) %  c->axis1)/(*Ray %  c->axis1);
  t[1] = ((c->base - *Org) % -c->axis1)/(*Ray % -c->axis1);
  t[2] = ((c->top  - *Org) %  c->axis2)/(*Ray %  c->axis2);
  t[3] = ((c->base - *Org) % -c->axis2)/(*Ray % -c->axis2);
  t[4] = ((c->top  - *Org) %  c->axis3)/(*Ray %  c->axis3);
  t[5] = ((c->base - *Org) % -c->axis3)/(*Ray % -c->axis3);

  char x_near = ((t[0]<t[1])? 0 : 1);
  char y_near = ((t[2]<t[3])? 2 : 3);
  char z_near = ((t[4]<t[5])? 4 : 5);

  RV_IntPtr->i = ((t[x_near] > t[y_near])? x_near : y_near); if(t[z_near] > t[RV_IntPtr->i]) RV_IntPtr->i=z_near;

  if( t[RV_IntPtr->i] < min(min(t[1-x_near],t[5-y_near]),t[9-z_near]) ){
	if(t[RV_IntPtr->i]<0.001f){
      RV_IntPtr->i = ((t[1-x_near] < t[5-y_near])? 1-x_near : 5-y_near); if(t[9-z_near] < t[RV_IntPtr->i]) RV_IntPtr->i=9-z_near;
	}
	RV_IntPtr->t = t[RV_IntPtr->i];
  } else {
	RV_IntPtr->t = -1.0f;
  }
}

void GetIPN(vector *Org, vector *Ray, IntPar *I)
{
  if(I->p->type < 2){
    if(((RV_TriPara *)(I->p))->sinfo && ((RV_TriPara *)(I->p))->sinfo->nrm){
      // phong interpolation
      I->N = unit(((RV_TriPara *)(I->p))->sinfo->nrm[0] + I->u*(((RV_TriPara *)(I->p))->sinfo->nrm[1] - ((RV_TriPara *)(I->p))->sinfo->nrm[0]) + I->v*(((RV_TriPara *)(I->p))->sinfo->nrm[2] - ((RV_TriPara *)(I->p))->sinfo->nrm[0]));
     } else {
      I->N  = ((RV_TriPara *)(I->p))->N;
     }
  } else if(I->p->type == 2){
    I->IP = *Org + I->t * *Ray;
    I->N  = unit(I->IP - ((RV_SimpleSphere *)(I->p))->position);
  } else if(I->p->type == 3){
    I->N  = unit(vector(I->IP.x, I->IP.y, (1.0f - I->IP.z)*((RV_Cylinder *)I->p)->r_top)*((RV_Cylinder *)I->p)->sinfo->NM);
    I->IP = *Org + I->t * *Ray;
  } else if(I->p->type == 4){
    I->IP = *Org + I->t * *Ray;
    I->N  = unit( (I->i<4)? ((I->i<2)? ((RV_Cube *)(I->p))->axis1: ((RV_Cube *)(I->p))->axis2) : ((RV_Cube *)(I->p))->axis3 );
    if((I->i) & 1) I->N = -(I->N);
  } else if(I->p->type == 5){
    I->N  = ((RV_SimpleRing *)(I->p))->N;
  } else if(I->p->type == 6){
    I->N  = unit(I->N * ((RV_Torus *)I->p)->sinfo->NM);
    I->IP = *Org + I->t * *Ray;
  } else if(I->p->type == 7){
    I->IP = *Org + I->t * *Ray;
    I->N  = unit(I->IP - ((RV_ComplexSphere *)(I->p))->position);
  } else if(I->p->type == 8){
    I->N  = unit(vector(2.0f*I->IP.x, 2.0f*I->IP.y, -1.0f) * ((RV_Paraboloid *)I->p)->sinfo->NM);
    I->IP = *Org + I->t * *Ray;
  } else if(I->p->type == 9){
    I->N  = unit(I->IP * ((RV_Ellipsoid *)I->p)->sinfo->NM);
    I->IP = *Org + I->t * *Ray;
  } else if(I->p->type == 10){
    I->N  = ((RV_ComplexRing *)(I->p))->N;
  }
}

void GetTXY(IntPar *I)
{
  if(I->p->type == 0){
    if(((RV_TriPara *)(I->p))->sinfo && ((RV_TriPara *)(I->p))->sinfo->txc){
	  I->x = ((RV_TriPara *)(I->p))->sinfo->txc[0].x + I->u*(((RV_TriPara *)(I->p))->sinfo->txc[1].x-((RV_TriPara *)(I->p))->sinfo->txc[0].x) + I->v*(((RV_TriPara *)(I->p))->sinfo->txc[2].x-((RV_TriPara *)(I->p))->sinfo->txc[0].x);
	  I->y = ((RV_TriPara *)(I->p))->sinfo->txc[0].y + I->u*(((RV_TriPara *)(I->p))->sinfo->txc[1].y-((RV_TriPara *)(I->p))->sinfo->txc[0].y) + I->v*(((RV_TriPara *)(I->p))->sinfo->txc[2].y-((RV_TriPara *)(I->p))->sinfo->txc[0].y);
	} else {
	  I->x = I->u;
	  I->y = I->v;
	}
  } else if(I->p->type == 1){
    if(((RV_TriPara *)(I->p))->sinfo && ((RV_TriPara *)(I->p))->sinfo->txc){
	  I->x = bilinear(I->u, I->v, ((RV_TriPara *)(I->p))->sinfo->txc[3].x, ((RV_TriPara *)(I->p))->sinfo->txc[1].x, ((RV_TriPara *)(I->p))->sinfo->txc[2].x, ((RV_TriPara *)(I->p))->sinfo->txc[0].x);
	  I->y = bilinear(I->u, I->v, ((RV_TriPara *)(I->p))->sinfo->txc[3].y, ((RV_TriPara *)(I->p))->sinfo->txc[1].y, ((RV_TriPara *)(I->p))->sinfo->txc[2].y, ((RV_TriPara *)(I->p))->sinfo->txc[0].y);
	} else {
	  I->x = I->u;
	  I->y = I->v;
	}
  } else if(I->p->type == 3){
	I->u /= ((RV_Cylinder *)I->p)->angle;
    if(((RV_Cylinder *)(I->p))->sinfo && ((RV_Cylinder *)(I->p))->sinfo->txc){
	  I->x = bilinear(I->v, I->u, ((RV_Cylinder *)(I->p))->sinfo->txc[3].x, ((RV_Cylinder *)(I->p))->sinfo->txc[2].x, ((RV_Cylinder *)(I->p))->sinfo->txc[1].x, ((RV_Cylinder *)(I->p))->sinfo->txc[0].x);
	  I->y = bilinear(I->v, I->u, ((RV_Cylinder *)(I->p))->sinfo->txc[3].y, ((RV_Cylinder *)(I->p))->sinfo->txc[2].y, ((RV_Cylinder *)(I->p))->sinfo->txc[1].y, ((RV_Cylinder *)(I->p))->sinfo->txc[0].y);
	} else {
	  I->x = I->u;
	  I->y = I->v;
    }
  } else if(I->p->type == 6){
	I->v /= (((RV_Torus *)I->p)->vangle? ((RV_Torus *)I->p)->vangle : 6.2831853f);
	if((((RV_Torus *)I->p)->ustart < 0.0f) && ((I->u-6.2831853f) >= ((RV_Torus *)I->p)->ustart)){
	  I->u = ((I->u-6.2831853f)-((RV_Torus *)I->p)->ustart)/(((RV_Torus *)I->p)->ustop-((RV_Torus *)I->p)->ustart);
	} else {
	  I->u = ( I->u            -((RV_Torus *)I->p)->ustart)/(((RV_Torus *)I->p)->ustop-((RV_Torus *)I->p)->ustart);
	}
    if(((RV_Torus *)(I->p))->sinfo && ((RV_Torus *)(I->p))->sinfo->txc){
	  I->x = bilinear(I->v, I->u, ((RV_Torus *)(I->p))->sinfo->txc[3].x, ((RV_Torus *)(I->p))->sinfo->txc[1].x, ((RV_Torus *)(I->p))->sinfo->txc[2].x, ((RV_Torus *)(I->p))->sinfo->txc[0].x);
	  I->y = bilinear(I->v, I->u, ((RV_Torus *)(I->p))->sinfo->txc[3].y, ((RV_Torus *)(I->p))->sinfo->txc[1].y, ((RV_Torus *)(I->p))->sinfo->txc[2].y, ((RV_Torus *)(I->p))->sinfo->txc[0].y);
	} else {
	  I->x = I->v;
	  I->y = I->u;
    }
  } else if(I->p->type == 7){
	I->v /= (((RV_ComplexSphere *)I->p)->vangle? ((RV_ComplexSphere *)I->p)->vangle : 6.2831853f);
    I->u  = (I->u-((RV_ComplexSphere *)I->p)->ustart)/(((RV_ComplexSphere *)I->p)->ustop-((RV_ComplexSphere *)I->p)->ustart);
    if(((RV_ComplexSphere *)(I->p))->sinfo && ((RV_ComplexSphere *)(I->p))->sinfo->txc){
	  I->x = bilinear(I->u, I->v, ((RV_ComplexSphere *)(I->p))->sinfo->txc[3].x, ((RV_ComplexSphere *)(I->p))->sinfo->txc[2].x, ((RV_ComplexSphere *)(I->p))->sinfo->txc[1].x, ((RV_ComplexSphere *)(I->p))->sinfo->txc[0].x);
	  I->y = bilinear(I->u, I->v, ((RV_ComplexSphere *)(I->p))->sinfo->txc[3].y, ((RV_ComplexSphere *)(I->p))->sinfo->txc[2].y, ((RV_ComplexSphere *)(I->p))->sinfo->txc[1].y, ((RV_ComplexSphere *)(I->p))->sinfo->txc[0].y);
	} else {
	  I->x = I->v;
	  I->y = I->u;
    }
  } else if(I->p->type == 8){
	I->v /= (((RV_Paraboloid *)I->p)->angle? ((RV_Paraboloid *)I->p)->angle : 6.2831853f);
    I->u  = sqrtf(I->u);
    if(((RV_Paraboloid *)(I->p))->sinfo && ((RV_Paraboloid *)(I->p))->sinfo->txc){
	  I->x = bilinear(I->u, I->v, ((RV_Paraboloid *)(I->p))->sinfo->txc[3].x, ((RV_Paraboloid *)(I->p))->sinfo->txc[2].x, ((RV_Paraboloid *)(I->p))->sinfo->txc[1].x, ((RV_Paraboloid *)(I->p))->sinfo->txc[0].x);
	  I->y = bilinear(I->u, I->v, ((RV_Paraboloid *)(I->p))->sinfo->txc[3].y, ((RV_Paraboloid *)(I->p))->sinfo->txc[2].y, ((RV_Paraboloid *)(I->p))->sinfo->txc[1].y, ((RV_Paraboloid *)(I->p))->sinfo->txc[0].y);
	} else {
	  I->x = I->v;
	  I->y = I->u;
    }
  } else if(I->p->type == 9){
	I->v /= (((RV_Ellipsoid *)I->p)->vangle? ((RV_Ellipsoid *)I->p)->vangle : 6.2831853f);
    I->u  = (I->u-((RV_Ellipsoid *)I->p)->ustart)/(((RV_Ellipsoid *)I->p)->ustop-((RV_Ellipsoid *)I->p)->ustart);
    if(((RV_Ellipsoid *)(I->p))->sinfo && ((RV_Ellipsoid *)(I->p))->sinfo->txc){
	  I->x = bilinear(I->u, I->v, ((RV_Ellipsoid *)(I->p))->sinfo->txc[3].x, ((RV_Ellipsoid *)(I->p))->sinfo->txc[2].x, ((RV_Ellipsoid *)(I->p))->sinfo->txc[1].x, ((RV_Ellipsoid *)(I->p))->sinfo->txc[0].x);
	  I->y = bilinear(I->u, I->v, ((RV_Ellipsoid *)(I->p))->sinfo->txc[3].y, ((RV_Ellipsoid *)(I->p))->sinfo->txc[2].y, ((RV_Ellipsoid *)(I->p))->sinfo->txc[1].y, ((RV_Ellipsoid *)(I->p))->sinfo->txc[0].y);
	} else {
	  I->x = I->v;
	  I->y = I->u;
    }
  } else if(I->p->type == 10){
	I->v /= (((RV_ComplexRing *)I->p)->angle? ((RV_ComplexRing *)I->p)->angle : 6.2831853f);
	float r_i=sqrtf(((RV_ComplexRing *)I->p)->r_i);
	float r_o=sqrtf(((RV_ComplexRing *)I->p)->r_o);
	I->u  = (sqrtf(I->u)-r_i)/(r_o-r_i);
    if(((RV_ComplexRing *)(I->p))->sinfo && ((RV_ComplexRing *)(I->p))->sinfo->txc){
	  I->x = bilinear(I->u, I->v, ((RV_ComplexRing *)(I->p))->sinfo->txc[3].x, ((RV_ComplexRing *)(I->p))->sinfo->txc[2].x, ((RV_ComplexRing *)(I->p))->sinfo->txc[1].x, ((RV_ComplexRing *)(I->p))->sinfo->txc[0].x);
	  I->y = bilinear(I->u, I->v, ((RV_ComplexRing *)(I->p))->sinfo->txc[3].y, ((RV_ComplexRing *)(I->p))->sinfo->txc[2].y, ((RV_ComplexRing *)(I->p))->sinfo->txc[1].y, ((RV_ComplexRing *)(I->p))->sinfo->txc[0].y);
	} else {
	  I->x = I->v;
	  I->y = I->u;
    }
  }
}

float R_fnf(float n1, float n2, double ce1);
float wlcoff(float cR, float cG, float cB, float v);
RGB_Color CallShade(RV_PrP OrgObj, MatPtr mat, vector *Org, vector *Ray, IntPar Cmin, float em, float dist, int level);
extern float wavlen;

RGB_Color ShadeDepth(vector *Org, vector *Ray, IntPar IntP, MatPtr mat, int level)
{
  GetIPN(Org, Ray, &IntP);
  float d = 1.0f-((distance(*Org,IntP.IP)-4.17f)/5.66f);
  return RGB_Color(d, d, d);
}

RGB_Color SIMD_ShNorm(vector *Org, vector Ray, float t, RV_PrP p)
{
  vector N;
  if(p->type<2){
    N=((RV_TriPara *)(p))->N;
  } else {
    vector IP = *Org + t*Ray; N  = unit(IP - ((RV_SimpleSphere *)p)->position);
  }
  return RGB_Color(0.5f*(1.0f+N.x), 0.5f*(1.0f+N.y), 0.5f*(1.0f+N.z));
}

RGB_Color ShadeNormals(vector *Org, vector *Ray, IntPar IntP, MatPtr mat, int level)
{
  GetIPN(Org, Ray, &IntP);
  return RGB_Color(0.5f*(1.0f+IntP.N.x), 0.5f*(1.0f+IntP.N.y), 0.5f*(1.0f+IntP.N.z));
}

RGB_Color ShadeReflection(vector *Org, vector *Ray, IntPar IntP, MatPtr mat, int level)
{
  GetIPN(Org, Ray, &IntP);
  vector L =  unit(RV_Lights->p - IntP.IP);
  vector R = -unit(2.0f*(IntP.N % L)*(IntP.N)-L);
  return RGB_Color(0.5f*(1.0f+R.x), 0.5f*(1.0f+R.y), 0.5f*(1.0f+R.z));
}

RGB_Color ShadeSimple(vector *Org, vector *Ray, IntPar IntP, MatPtr mat, int level)
{
  GetIPN(Org, Ray, &IntP);
  vector L = unit(RV_Lights->p - IntP.IP);
  float  s = (IntP.p->flags & RV_ReceiveNoShadow)? 1.0f : 1.0f-ShadowTest(IntP.p, RV_Lights->Prim, &(IntP.IP), &L, distance(RV_Lights->p, IntP.IP));
  float NL; if(s){ NL = s*(L % IntP.N); if(NL<0.0f) NL=0.0f; } else NL=0.0f;
  RGB_Color C; C.R = (Ambient.R + NL) * IntP.p->Mat->Reflection.R;
               C.G = (Ambient.G + NL) * IntP.p->Mat->Reflection.G;
               C.B = (Ambient.B + NL) * IntP.p->Mat->Reflection.B;
  return C;
}

extern float wrap(float c);

RGB_Color ShadeUV(vector *Org, vector *Ray, IntPar IntP, MatPtr mat, int level)
{
  GetIPN(Org, Ray, &IntP); GetTXY(&IntP);
  return RGB_Color(wrap(IntP.x), wrap(IntP.y), 0.5f);
}

RGB_Color Shade(vector *Org, vector *Ray, IntPar IntP, MatPtr mat, int level)
{
  const int max_raydepth=3;

  GetIPN(Org, Ray, &IntP); MatPtr m = IntP.p->Mat;
  float NL=0.0f,RV=0.0f; RGB_Color C;
  LitPtr l=RV_Lights;
  while(l){
    vector L = unit(l->p - IntP.IP);
    float  s = (IntP.p->flags & RV_ReceiveNoShadow)? 1.0f : 1.0f-ShadowTest(IntP.p, l->Prim, &(IntP.IP), &L, distance(l->p, IntP.IP));
    if(s){
      float NL1 = L % IntP.N;
      if(level==0){
        vector R  = -unit(2.0f*NL1*(IntP.N)-L);
        float RV1 = R % unit(IntP.IP - *Org); if(RV1<0.0f) RV1=0.0f; RV1 = (float)pow(RV1, (m->PBExp? m->PBExp : 30.0));
        RV += s * l->v * RV1;
      }
      if(NL1>0.0f) NL += s * l->v * NL1;      
    }
    l=l->next;
  }
  if(m->Texture){
    GetTXY(&IntP);
    if(m->Texture->tbits){
      unsigned char *bitptr = m->Texture->tbits + ((int)(wrap(IntP.y)*m->Texture->tinfo.bmiHeader.biHeight)) * ((m->Texture->tinfo.bmiHeader.biWidth*((m->Texture->tinfo.bmiHeader.biBitCount)>>3)+3) & 0xFFFFFFFC);
      unsigned long  offset = 3*                  ((int)(wrap(IntP.x)*m->Texture->tinfo.bmiHeader.biWidth));
      C.R = (bitptr[offset+2]/255.0f); C.G = (bitptr[offset+1]/255.0f); C.B = (bitptr[offset]/255.0f);
    } else {
      C = RGB_Color(wrap(IntP.x), wrap(IntP.y), 0.5f);
    }
  } else {
    if((IntP.p->flags & RV_Chequered) && (((int)ceil(2.0f*IntP.IP.x)^(int)ceil(2.0f*IntP.IP.y)) & 1)){
      C = RGB_Color(0.8f, 0.8f, 0.8f);
    } else {
      if(m->Reflection.G<0.0f){
        C = RGB_Color(m->Reflection.R, m->Reflection.R, m->Reflection.R);
      } else {
        C = m->Reflection;
      }
    }
  }
  C.R = (Ambient.R + NL) * (1.0f - m->Specularity) * C.R + m->Specularity*RV;
  C.G = (Ambient.G + NL) * (1.0f - m->Specularity) * C.G + m->Specularity*RV;
  C.B = (Ambient.B + NL) * (1.0f - m->Specularity) * C.B + m->Specularity*RV;

  // specular reflection/refraction
  if((level<max_raydepth) && (m->Specularity>ray_thresh)){
    float  spc = 0.0f, trn=0.0f;
    float  ca  = IntP.N % *Ray;
    vector OutDir;

    if(m->Transmission.R < 0.0f){
      // fresnel reflection?
      float n1,n2;
      if(ca<0.0f){
        n1=mat->IOR(wavlen); n2=            m->IOR(wavlen);
      } else {
        n1=m  ->IOR(wavlen); n2=RV_Environment.IOR(wavlen);
      }
      spc = m->Specularity * R_fnf(n1,n2,ca); trn = m->Specularity-spc;
    } else {
      // simple specular reflection/refraction?
      if((m->Specularity > ray_thresh) && (m->type!=Shader_Phong)){
        spc = m->Specularity * ((m->Reflection.G < 0.0f)? m->Reflection.R : wlcoff(m->Reflection.R, m->Reflection.G, m->Reflection.B, wavlen));
        trn = m->Transmission.R;
      }
    }

    // collect specular reflected light
    if(spc > ray_thresh) {
      // mirror direction
      OutDir = -unit(2.0f * ca * IntP.N - *Ray); //if(m->PBExp > 0.001f) OutDir = rndpdir(OutDir, 1.0f/m->PBExp);
      C += spc * RayTrace(IntP.p, NULL, &RV_Environment, &(IntP.IP), &OutDir, 1.0f, 0.0f, &CallShade, level+1, max_raydepth);
    }
    // collect transmitted (refracted) light
    if(trn > ray_thresh) {
      if(ca<0.0f){
        OutDir=refract(*Ray,  IntP.N, mat->IOR(wavlen),             m->IOR(wavlen));
        C += trn * RayTrace(IntP.p, NULL,                           m                    , &(IntP.IP), &OutDir, 1.0f, 0.0f, &CallShade, level+1, max_raydepth);
      } else {
        OutDir=refract(*Ray, -IntP.N, mat->IOR(wavlen), RV_Environment.IOR(wavlen));
        C += trn * RayTrace(IntP.p, NULL, ((OutDir % IntP.N)<0.0f)? m : &(RV_Environment), &(IntP.IP), &OutDir, 1.0f, 0.0f, &CallShade, level+1, max_raydepth);
      }
    }
  }
  return C;
}

RGB_Color (*RV_Shader)(vector *Org, vector *Ray, IntPar IntP, MatPtr mat, int level) = &Shade;

RGB_Color CallShade(RV_PrP OrgObj, MatPtr mat, vector *Org, vector *Ray, IntPar Cmin, float em, float dist, int level)
{
  return RV_Shader(Org, Ray, Cmin, mat, level);
}

extern "C" {
  void RV_DispatchIntersectOther0(RV_PrP p)
  {
	//if((p->flags & RV_OnlyIndirect){ RV_IntPtr->t = -1.0f; return; }
           if(p->type== 3){
	  RV_IntersectCylinder0((RV_Cylinder *)p);
	} else if(p->type== 5){
	  RV_IntersectSimpleRing0((RV_SimpleRing *)p);
	} else if(p->type== 6){
	  RV_IntersectTorus0((RV_Torus *)p);
	} else if(p->type== 7){
	  RV_IntersectComplexSphere0((RV_ComplexSphere *)p);
	} else if(p->type== 8){
	  RV_IntersectParaboloid0((RV_Paraboloid *)p);
	} else if(p->type== 9){
	  RV_IntersectEllipsoid0((RV_Ellipsoid *)p);
	} else if(p->type==10){
	  RV_IntersectComplexRing0((RV_ComplexRing *)p);
	} else {
      RV_IntPtr->t = -1.0f;
	}
  }
  void RV_DispatchIntersectAll0(RV_PrP p)
  {
	if(p->lastray == cray){ RV_IntPtr->t = -1.0f; return; } else p->lastray = cray;
    if((p->type==1) || (p->type==0)){
      RV_IntersectTriPara0((RV_TriPara *)p);
	} else if(p->type==2){
	  RV_IntersectSimpleSphere0((RV_SimpleSphere *)p);
	} else if(p->type==4){
	  RV_IntersectCube0((RV_Cube *)p);
	} else {
      RV_DispatchIntersectOther0(p);
	}
  }
  void RV_DispatchIntersectOther1(RV_PrP p)
  {
           if(p->type== 3){
	  RV_IntersectCylinder1((RV_Cylinder *)p, RV_OrgPtr, RV_DirPtr);
	} else if(p->type== 5){
	  RV_IntersectSimpleRing1((RV_SimpleRing *)p, RV_OrgPtr, RV_DirPtr);
	} else if(p->type== 6){
	  RV_IntersectTorus1((RV_Torus *)p, RV_OrgPtr, RV_DirPtr);
	} else if(p->type== 7){
	  RV_IntersectComplexSphere1((RV_ComplexSphere *)p, RV_OrgPtr, RV_DirPtr);
	} else if(p->type== 8){
	  RV_IntersectParaboloid1((RV_Paraboloid *)p, RV_OrgPtr, RV_DirPtr);
	} else if(p->type== 9){
	  RV_IntersectEllipsoid1((RV_Ellipsoid *)p, RV_OrgPtr, RV_DirPtr);
	} else if(p->type==10){
	  RV_IntersectComplexRing1((RV_ComplexRing *)p, RV_OrgPtr, RV_DirPtr);
	} else {
      RV_IntPtr->t = -1.0f;
	}
  }
  void RV_DispatchIntersectAll1(RV_PrP p)
  {
	if(p->lastray == cray){ RV_IntPtr->t = -1.0f; return; } else p->lastray = cray;
    if((p->type==1) || (p->type==0)){
      RV_IntersectTriPara1((RV_TriPara *)p, RV_OrgPtr, RV_DirPtr);
	} else if(p->type==2){
	  RV_IntersectSimpleSphere1((RV_SimpleSphere *)p, RV_OrgPtr, RV_DirPtr);
	} else if(p->type==4){
	  RV_IntersectCube1((RV_Cube *)p, RV_OrgPtr, RV_DirPtr);
	} else {
      RV_DispatchIntersectOther1(p);
	}
  }
}

void TraceTree0(BSPplane *tree, vector HP)
{
  float olm;
  switch(tree->axis){
    case 0:
      if(UseSSE) RV_TraceList0((PrLP)tree->left);
      else {
        PrLP p = (PrLP)tree->left;
        while(p){
          RV_DispatchIntersectAll0(p->Prim);
          if(RV_IntPtr->t>=0.0f){
            if((Cmin.t<0.0f) || (RV_IntPtr->t<Cmin.t)){ RV_IntPtr->p=p->Prim; Cmin=*RV_IntPtr; }
          }
          p=p->next;
        }
      }
    break;
    case 1:
      if(HP.x < tree->dval){
        olm=BBMax.x; BBMax.x=tree->dval; TraceTree0(tree->left, HP); BBMax.x=olm;
        if((Cmin.t>=0.0f) && (HP.x > Cmin.IP.x)) return;
        if(RV_RayPtr->x > 0.0f){
          vector PH = *RV_CamPtr + ((tree->dval - RV_CamPtr->x)/RV_RayPtr->x) * *RV_RayPtr;
          if((PH.x>BBMin.x)&&(PH.x<BBMax.x)&&(PH.y>BBMin.y)&&(PH.y<BBMax.y)&&(PH.z>BBMin.z)&&(PH.z<BBMax.z)){
            olm=BBMin.x; BBMin.x=tree->dval; TraceTree0(tree->rigt, PH); BBMin.x=olm;
          }
        }
      } else {
        olm=BBMin.x; BBMin.x=tree->dval; TraceTree0(tree->rigt, HP); BBMin.x=olm;
        if((Cmin.t>=0.0f) && (HP.x < Cmin.IP.x)) return;
        if(RV_RayPtr->x < 0.0f){
          vector PH = *RV_CamPtr + ((tree->dval - RV_CamPtr->x)/RV_RayPtr->x) * *RV_RayPtr;
          if((PH.x>BBMin.x)&&(PH.x<BBMax.x)&&(PH.y>BBMin.y)&&(PH.y<BBMax.y)&&(PH.z>BBMin.z)&&(PH.z<BBMax.z)){
            olm=BBMax.x; BBMax.x=tree->dval; TraceTree0(tree->left, PH); BBMax.x=olm;
          }
        }
      }
    break;
    case 2:
      if(HP.y < tree->dval){
        olm=BBMax.y; BBMax.y=tree->dval; TraceTree0(tree->left, HP); BBMax.y=olm;
        if((Cmin.t>=0.0f) && (HP.y > Cmin.IP.y)) return;
        if(RV_RayPtr->y > 0.0f){
          vector PH = *RV_CamPtr + ((tree->dval - RV_CamPtr->y)/RV_RayPtr->y) * *RV_RayPtr;
          if((PH.x>BBMin.x)&&(PH.x<BBMax.x)&&(PH.y>BBMin.y)&&(PH.y<BBMax.y)&&(PH.z>BBMin.z)&&(PH.z<BBMax.z)){
	        olm=BBMin.y; BBMin.y=tree->dval; TraceTree0(tree->rigt, PH); BBMin.y=olm;
		  }
		}
	  } else {
	    olm=BBMin.y; BBMin.y=tree->dval; TraceTree0(tree->rigt, HP); BBMin.y=olm;
	    if((Cmin.t>=0.0f) && (HP.y < Cmin.IP.y)) return;
        if(RV_RayPtr->y < 0.0f){
          vector PH = *RV_CamPtr + ((tree->dval - RV_CamPtr->y)/RV_RayPtr->y) * *RV_RayPtr;
          if((PH.x>BBMin.x)&&(PH.x<BBMax.x)&&(PH.y>BBMin.y)&&(PH.y<BBMax.y)&&(PH.z>BBMin.z)&&(PH.z<BBMax.z)){
	        olm=BBMax.y; BBMax.y=tree->dval; TraceTree0(tree->left, PH); BBMax.y=olm;
		  }
		}
	  }
	break;
	case 3:
	  if(HP.z < tree->dval){
	    olm=BBMax.z; BBMax.z=tree->dval; TraceTree0(tree->left, HP); BBMax.z=olm;
	    if((Cmin.t>=0.0f) && (HP.z > Cmin.IP.z)) return;
        if(RV_RayPtr->z > 0.0f){
          vector PH = *RV_CamPtr + ((tree->dval - RV_CamPtr->z)/RV_RayPtr->z) * *RV_RayPtr;
          if((PH.x>BBMin.x)&&(PH.x<BBMax.x)&&(PH.y>BBMin.y)&&(PH.y<BBMax.y)&&(PH.z>BBMin.z)&&(PH.z<BBMax.z)){
	        olm=BBMin.z; BBMin.z=tree->dval; TraceTree0(tree->rigt, PH); BBMin.z=olm;
		  }
		}
	  } else {
	    olm=BBMin.z; BBMin.z=tree->dval; TraceTree0(tree->rigt, HP); BBMin.z=olm;
	    if((Cmin.t>=0.0f) && (HP.z < Cmin.IP.z)) return;
        if(RV_RayPtr->z < 0.0f){
          vector PH = *RV_CamPtr + ((tree->dval - RV_CamPtr->z)/RV_RayPtr->z) * *RV_RayPtr;
          if((PH.x>BBMin.x)&&(PH.x<BBMax.x)&&(PH.y>BBMin.y)&&(PH.y<BBMax.y)&&(PH.z>BBMin.z)&&(PH.z<BBMax.z)){
	        olm=BBMax.z; BBMax.z=tree->dval; TraceTree0(tree->left, PH); BBMax.z=olm;
		  }
		}
	  }
	break;
  }
}

void TraceTree1(BSPplane *tree, vector HP)
{
  float olm;
  switch(tree->axis){
    case 0:
	  if(UseSSE) RV_TraceList1((PrLP)tree->left);
	  else {
        PrLP p = (PrLP)tree->left;
        while(p){
	      if(!(*RV_ShfPtr) || !(p->Prim->flags & RV_CastNoShadow) ){
            RV_DispatchIntersectAll1(p->Prim);
            if((p->Prim == OrgOb) && (RV_IntPtr->t<0.0001f)) RV_IntPtr->t=-1.0f;
		    if(*RV_ShfPtr && (RV_IntPtr->t > *RV_ShfPtr))    RV_IntPtr->t=-1.0f;
		    if(*RV_ShfPtr && (p->Prim == CLtOb))             RV_IntPtr->t=-1.0f;
            if(RV_IntPtr->t>=0.0f){
              if((Cmin.t<0.0f) || (RV_IntPtr->t<Cmin.t)){ RV_IntPtr->p=p->Prim; Cmin=*RV_IntPtr; }
              if(*RV_ShfPtr) return;
			}
		  }
         p=p->next;
		}
	  }
	break;
	case 1:
	  if(HP.x < tree->dval){
	    olm=BBMax.x; BBMax.x=tree->dval; TraceTree1(tree->left, HP); BBMax.x=olm;
	    if((Cmin.t>=0.0f) && ((HP.x > Cmin.IP.x) || *RV_ShfPtr)) return;
        if(RV_DirPtr->x > 0.0f){
          vector PH = *RV_OrgPtr + ((tree->dval - RV_OrgPtr->x)/RV_DirPtr->x) * *RV_DirPtr;
          if((PH.x>BBMin.x)&&(PH.x<BBMax.x)&&(PH.y>BBMin.y)&&(PH.y<BBMax.y)&&(PH.z>BBMin.z)&&(PH.z<BBMax.z)){
	        olm=BBMin.x; BBMin.x=tree->dval; TraceTree1(tree->rigt, PH); BBMin.x=olm;
		  }
		}
	  } else {
	    olm=BBMin.x; BBMin.x=tree->dval; TraceTree1(tree->rigt, HP); BBMin.x=olm;
	    if((Cmin.t>=0.0f) && ((HP.x < Cmin.IP.x) || *RV_ShfPtr)) return;
        if(RV_DirPtr->x < 0.0f){
          vector PH = *RV_OrgPtr + ((tree->dval - RV_OrgPtr->x)/RV_DirPtr->x) * *RV_DirPtr;
          if((PH.x>BBMin.x)&&(PH.x<BBMax.x)&&(PH.y>BBMin.y)&&(PH.y<BBMax.y)&&(PH.z>BBMin.z)&&(PH.z<BBMax.z)){
	        olm=BBMax.x; BBMax.x=tree->dval; TraceTree1(tree->left, PH); BBMax.x=olm;
		  }
		}
	  }
	break;
	case 2:
	  if(HP.y < tree->dval){
	    olm=BBMax.y; BBMax.y=tree->dval; TraceTree1(tree->left, HP); BBMax.y=olm;
	    if((Cmin.t>=0.0f) && ((HP.y > Cmin.IP.y) || *RV_ShfPtr)) return;
        if(RV_DirPtr->y > 0.0f){
          vector PH = *RV_OrgPtr + ((tree->dval - RV_OrgPtr->y)/RV_DirPtr->y) * *RV_DirPtr;
          if((PH.x>BBMin.x)&&(PH.x<BBMax.x)&&(PH.y>BBMin.y)&&(PH.y<BBMax.y)&&(PH.z>BBMin.z)&&(PH.z<BBMax.z)){
	        olm=BBMin.y; BBMin.y=tree->dval; TraceTree1(tree->rigt, PH); BBMin.y=olm;
		  }
		}
	  } else {
	    olm=BBMin.y; BBMin.y=tree->dval; TraceTree1(tree->rigt, HP); BBMin.y=olm;
	    if((Cmin.t>=0.0f) && ((HP.y < Cmin.IP.y) || *RV_ShfPtr)) return;
        if(RV_DirPtr->y < 0.0f){
          vector PH = *RV_OrgPtr + ((tree->dval - RV_OrgPtr->y)/RV_DirPtr->y) * *RV_DirPtr;
          if((PH.x>BBMin.x)&&(PH.x<BBMax.x)&&(PH.y>BBMin.y)&&(PH.y<BBMax.y)&&(PH.z>BBMin.z)&&(PH.z<BBMax.z)){
	        olm=BBMax.y; BBMax.y=tree->dval; TraceTree1(tree->left, PH); BBMax.y=olm;
		  }
		}
	  }
	break;
	case 3:
	  if(HP.z < tree->dval){
	    olm=BBMax.z; BBMax.z=tree->dval; TraceTree1(tree->left, HP); BBMax.z=olm;
	    if((Cmin.t>=0.0f) && ((HP.z > Cmin.IP.z) || *RV_ShfPtr)) return;
        if(RV_DirPtr->z > 0.0f){
          vector PH = *RV_OrgPtr + ((tree->dval - RV_OrgPtr->z)/RV_DirPtr->z) * *RV_DirPtr;
          if((PH.x>BBMin.x)&&(PH.x<BBMax.x)&&(PH.y>BBMin.y)&&(PH.y<BBMax.y)&&(PH.z>BBMin.z)&&(PH.z<BBMax.z)){
	        olm=BBMin.z; BBMin.z=tree->dval; TraceTree1(tree->rigt, PH); BBMin.z=olm;
		  }
		}
	  } else {
	    olm=BBMin.z; BBMin.z=tree->dval; TraceTree1(tree->rigt, HP); BBMin.z=olm;
	    if((Cmin.t>=0.0f) && ((HP.z < Cmin.IP.z) || *RV_ShfPtr)) return;
        if(RV_DirPtr->z < 0.0f){
          vector PH = *RV_OrgPtr + ((tree->dval - RV_OrgPtr->z)/RV_DirPtr->z) * *RV_DirPtr;
          if((PH.x>BBMin.x)&&(PH.x<BBMax.x)&&(PH.y>BBMin.y)&&(PH.y<BBMax.y)&&(PH.z>BBMin.z)&&(PH.z<BBMax.z)){
	        olm=BBMax.z; BBMax.z=tree->dval; TraceTree1(tree->left, PH); BBMax.z=olm;
		  }
		}
	  }
	break;
  }
}

float DivQuad(float xb, float yb, float ll, vector *Org, float v1, float v2, float v3, float v4, int level)
{
  vector TR;
  float v0,v12,v23,v34,v14;
  if(fabs(v2-v1)>0.1f){
    TR = ((RV_TriPara *)(CLtOb))->base + (xb + 0.5f*ll)*((RV_TriPara *)(CLtOb))->edg1 + (yb          )*((RV_TriPara *)(CLtOb))->edg2 - *Org; *RV_ShfPtr = Betrag(TR); *RV_DirPtr = TR/(*RV_ShfPtr); *RV_ShfPtr *= 0.999f;
    Cmin.t=-1.0f; cray++; TraceTree1(RV_SceneTree, *RV_OrgPtr); v12 = (Cmin.t>0.0f)? 1.0f : 0.0f;
  } else v12=0.5f*(v1+v2);
  if(fabs(v3-v4)>0.1f){
    TR = ((RV_TriPara *)(CLtOb))->base + (xb + 0.5f*ll)*((RV_TriPara *)(CLtOb))->edg1 + (yb +      ll)*((RV_TriPara *)(CLtOb))->edg2 - *Org; *RV_ShfPtr = Betrag(TR); *RV_DirPtr = TR/(*RV_ShfPtr); *RV_ShfPtr *= 0.999f;
    Cmin.t=-1.0f; cray++; TraceTree1(RV_SceneTree, *RV_OrgPtr); v34 = (Cmin.t>0.0f)? 1.0f : 0.0f;
  } else v34=0.5f*(v3+v4);
  if(fabs(v3-v2)>0.1f){
    TR = ((RV_TriPara *)(CLtOb))->base + (xb +      ll)*((RV_TriPara *)(CLtOb))->edg1 + (yb + 0.5f*ll)*((RV_TriPara *)(CLtOb))->edg2 - *Org; *RV_ShfPtr = Betrag(TR); *RV_DirPtr = TR/(*RV_ShfPtr); *RV_ShfPtr *= 0.999f;
    Cmin.t=-1.0f; cray++; TraceTree1(RV_SceneTree, *RV_OrgPtr); v23 = (Cmin.t>0.0f)? 1.0f : 0.0f;
  } else v23=0.5f*(v3+v2);
  if(fabs(v4-v1)>0.1f){
    TR = ((RV_TriPara *)(CLtOb))->base + (xb          )*((RV_TriPara *)(CLtOb))->edg1 + (yb + 0.5f*ll)*((RV_TriPara *)(CLtOb))->edg2 - *Org; *RV_ShfPtr = Betrag(TR); *RV_DirPtr = TR/(*RV_ShfPtr); *RV_ShfPtr *= 0.999f;
    Cmin.t=-1.0f; cray++; TraceTree1(RV_SceneTree, *RV_OrgPtr); v14 = (Cmin.t>0.0f)? 1.0f : 0.0f;
  } else v14=0.5f*(v1+v4);
  if((fabs(v12-v34)>0.1f) || (fabs(v23-v14)>0.1f)){
    TR = ((RV_TriPara *)(CLtOb))->base + (xb + 0.5f*ll)*((RV_TriPara *)(CLtOb))->edg1 + (yb + 0.5f*ll)*((RV_TriPara *)(CLtOb))->edg2 - *Org; *RV_ShfPtr = Betrag(TR); *RV_DirPtr = TR/(*RV_ShfPtr); *RV_ShfPtr *= 0.999f;
    Cmin.t=-1.0f; cray++; TraceTree1(RV_SceneTree, *RV_OrgPtr); v0 = (Cmin.t>0.0f)? 1.0f : 0.0f;
  } else v0=0.25f*(v12+v34+v23+v14);
  if(level<2){
	float q1=DivQuad(xb        ,yb        ,0.5f*ll,Org,v1,v12,v0,v14,level+1);
	float q2=DivQuad(xb+0.5f*ll,yb        ,0.5f*ll,Org,v12,v2,v23,v0,level+1);
	float q3=DivQuad(xb+0.5f*ll,yb+0.5f*ll,0.5f*ll,Org,v0,v23,v3,v34,level+1);
	float q4=DivQuad(xb        ,yb+0.5f*ll,0.5f*ll,Org,v14,v0,v34,v4,level+1);
    return (q1+q2+q3+q4)/4.0f;
  }
  return (v14+v1+v12+v2+v23+v3+v34+v4+v0)/9.0f;
}

float ShadowTest(RV_PrP OrgObj, RV_PrP LtObj, vector *Org, vector *Ray, float dist)
{
  *RV_OrgPtr = *Org; OrgOb=OrgObj; CLtOb=LtObj;
  if(LtObj && (LtObj->type==1)){
    vector TR;
    TR = ((RV_TriPara *)(CLtOb))->base                                                                 - *Org; *RV_ShfPtr = Betrag(TR); *RV_DirPtr = TR/(*RV_ShfPtr); *RV_ShfPtr *= 0.999f;
    Cmin.t=-1.0f; cray++; TraceTree1(RV_SceneTree, *RV_OrgPtr); float v1 = (Cmin.t>0.0f)? 1.0f : 0.0f;
    TR = ((RV_TriPara *)(CLtOb))->base + ((RV_TriPara *)(CLtOb))->edg1                                 - *Org; *RV_ShfPtr = Betrag(TR); *RV_DirPtr = TR/(*RV_ShfPtr); *RV_ShfPtr *= 0.999f;
    Cmin.t=-1.0f; cray++; TraceTree1(RV_SceneTree, *RV_OrgPtr); float v2 = (Cmin.t>0.0f)? 1.0f : 0.0f;
    TR = ((RV_TriPara *)(CLtOb))->base + ((RV_TriPara *)(CLtOb))->edg1 + ((RV_TriPara *)(CLtOb))->edg2 - *Org; *RV_ShfPtr = Betrag(TR); *RV_DirPtr = TR/(*RV_ShfPtr); *RV_ShfPtr *= 0.999f;
    Cmin.t=-1.0f; cray++; TraceTree1(RV_SceneTree, *RV_OrgPtr); float v3 = (Cmin.t>0.0f)? 1.0f : 0.0f;
    TR = ((RV_TriPara *)(CLtOb))->base                                 + ((RV_TriPara *)(CLtOb))->edg2 - *Org; *RV_ShfPtr = Betrag(TR); *RV_DirPtr = TR/(*RV_ShfPtr); *RV_ShfPtr *= 0.999f;
    Cmin.t=-1.0f; cray++; TraceTree1(RV_SceneTree, *RV_OrgPtr); float v4 = (Cmin.t>0.0f)? 1.0f : 0.0f;
    return DivQuad(0.0f, 0.0f, 1.0f, Org, v1, v2, v3, v4, 0);
  } else {
    *RV_DirPtr = *Ray; *RV_ShfPtr = 0.999f*dist; Cmin.t=-1.0f; cray++; TraceTree1(RV_SceneTree, *RV_OrgPtr);
    if(Cmin.t>0.0f) return 1.0f; else return 0.0f;
  }
}

void RayTrace4(void)
{
  RV_MDP[52]=-1.0f; RV_MDP[53]=-1.0f; RV_MDP[54]=-1.0f; RV_MDP[55]=-1.0f;

  RV_SIMD_TrLst0((PrLP)RV_SceneTree->left);
}

RGB_Color RayTrace0(vector RayVec)
{
  RGB_Color col=RGB_Color(0.0f, 0.0f, 0.0f); // default return value

  *RV_RayPtr=RayVec; Cmin.t=-1.0f; cray++;

  if(RV_LtObj){ RV_DispatchIntersectAll0(RV_LtObj); RV_IntPtr->p=RV_LtObj; Cmin=*RV_IntPtr; }
  
  if(UseSSE) RV_IntersectBoundingBox0(); else RV_IntersectBoundingBox0c();
  if(RV_IntPtr->t>0.0f) TraceTree0(RV_SceneTree, RV_IntPtr->IP);

  if(Cmin.t>0.0f){
    col=RV_Shader(RV_CamPtr, RV_RayPtr, Cmin, &RV_Environment, 0);
  }

  return col;
  //return RayTrace(NULL, NULL, &RV_Environment, RV_CamPtr, &RayVec, 1.0f, 0.0f, &CallShade, 0, 1);
  //return RGB_Color(fabs(RayVec.x),fabs(RayVec.y),fabs(RayVec.z));
}

RGB_Color RayTrace(RV_PrP    OrgObj,           // object where ray comes from
                   void      Init(vector Org), // initialization on ray origin
                   MatPtr    mat,              // material where ray is in
                   vector   *Org,              // start-position of ray
                   vector   *Ray,              // direction of ray
                   float     em,               // energy carried by ray
                   float     dist,             // accumulated distance from ray emitter
                   RGB_Color shade(RV_PrP OrgObj, MatPtr mat, vector *Org, vector *Ray, IntPar Cmin, float em, float dist, int level), // shader function
                   int       level,            // current recursion level
                   int       max_raydepth)     // maximum recursion level
{
  Cmin.t=-1.0f; cray++;

  if(Init!=NULL) Init(*Org);      // ray initialization
  
  RGB_Color col = RGB_Color(0.0f, 0.0f, 0.0f); // default return value (background color)

  *RV_OrgPtr = *Org; *RV_DirPtr = *Ray; OrgOb=OrgObj; *RV_ShfPtr = 0.0f;
  
  if(RV_LtObj) { RV_DispatchIntersectAll1(RV_LtObj); RV_IntPtr->p=RV_LtObj; Cmin=*RV_IntPtr; }

  RV_IntersectBoundingBox1(); if(RV_IntPtr->t>0.0f) TraceTree1(RV_SceneTree, RV_IntPtr->IP);

  if(Cmin.t>0.0f){
    // diffuse shading at intersection point
	col  = shade(OrgObj, mat, Org, Ray, Cmin, em, dist, level);
	// absorbtion
	//col *= mat->Thru(wavlen, dt); 
  }
  return col;
}
